/*
 * ----------------------------------------------
 *
 * This software belongs to the zzzutils package.
 *
 * Copyright (c) 2003 Fabian "zzznowman" Pietsch.
 * All rights reserved.
 *
 * ----------------------------------------------
 */

/*
 * zbtree.c -- provide an unbalanced b-tree implementation for zzzutils
 *
 **********************************************************************
 *
 * brief ChangeLog:
 *   2003-08-06 to -08 first implementation inside of zzzupdate-xmms.c
 *   2003-08-08 abstraction as zbtree.{c,h}
 *
 **********************************************************************
 *
 *
 */

#include "zzzutils.h"
#include "zbtree.h"

/* required for malloc() */
#include <stdlib.h>

/* required for strcoll() */
#include <string.h>


/* zbtree_add -- add a new element to a struct zbtree chain */
struct zbtree *zbtree_add(struct zbtree **bt, char *v)
{
	struct zbtree  *bt_new,
	              **bt_cur = bt;
	
	/* create new zbtree element */
	if ((bt_new = (struct zbtree *) calloc(1, sizeof(struct zbtree))) == NULL) {
		zperror("zbtree_add(): malloc()");
		return NULL;
	}
	bt_new->v = v;

	/* search free hook in bt to link bt_new into */
	while (*bt_cur != NULL) {
		/* compare new element's value to that of the current element's */
		if (strcoll(v, (**bt_cur).v) > 0)
			bt_cur = &(**bt_cur).n_gt;
		else
			bt_cur = &(**bt_cur).n_le;
	}

	/* link bt_new into free hook in bt */
	*bt_cur = bt_new;
	
	/* return pointer to added element */
	return bt_new;
}

/* zbtree_free -- free an entire struct zbtree chain */
void zbtree_free(struct zbtree **bt, const int freev)
{
	/* invalid / *NULL pointer? */
	if (!*bt) {
#if defined(DEBUG)
		zwarn("zbtree_free(): not freeing *NULL pointer");
#endif
		return;
	}

	/* subtrees to free? */
	if ((**bt).n_gt != NULL)
		zbtree_free(&(**bt).n_gt, freev);
	if ((**bt).n_le != NULL)
		zbtree_free(&(**bt).n_le, freev);

	/* also free contained data? */
	if (freev)
		(**bt).v = realloc((**bt).v, 0);

	/* finally free zbtree element itself */
	*bt = realloc(*bt, 0);
}


/*
 * zbtree_e*() -- zbtree enumeration functions
 ***********************************************************************
 * zbtree_estart() has to be called first with the zbtree's root element
 * zbtree_enext() is called then to sequentially get the tree's elements
 * zbtree_efirst() is a wrapper around zbtree_enext():
 *     it resets the enumeration & returns the zbtree's first element
 * zbtree_ereset() resets the enumeration, so ending and starting again
 *     the same enumeration becomes unneccessarry
 * zbtree_eend() has to be called at the end of every enumeration, espe-
 *     cially if only partial -- it frees the data structures used
 */

/* zbtree_estart -- start enumeration */
struct zbtree_enum *zbtree_estart(struct zbtree *bt)
{
	struct zbtree_enum *e;

	/* allocating e with calloc() is important:
	 * it explicitly clears the allocated memory
	 */
	if ((e = (struct zbtree_enum *) calloc(1, sizeof(struct zbtree_enum)))
	    == NULL ){
		zperror("zbtree_estart(): malloc()");
		return NULL;
	}
	e->root = bt;
	return e;
}

/* zbtree_efirst -- reset zbtree_enum and return first element in the tree */
struct zbtree *zbtree_efirst(struct zbtree_enum *e)
{
	if (!e) {
		zerror("zbtree_efirst(): called with NULL pointer");
		return NULL;
	}
	if (e->stack) {
		e->stack = realloc(e->stack, 0);
		e->stack_size = 0;
	}
	e->stack_depth = 0;
	return zbtree_enext(e);
}

/* zbtree_enext -- return next element in zbtree enumeration */
struct zbtree *zbtree_enext(struct zbtree_enum *e)
{
	struct zbtree *bt_cur, *bt_next;
	
	if (!e) {
		zerror("zbtree_enext(): called with NULL pointer");
		return NULL;
	}
	/* begin of enumeration? */
	/* if ((e->stack == NULL) || (e->stack[0] != bt)) { */
	if (e->stack == NULL) {
#ifdef DEBUG
		zwarn("zbtree_enext(): new enumeration");
#endif
		/* if (e->stack == NULL) { */
		e->stack_size = 0;
		/* } */
		e->stack_depth = 0;

		/* first call, so start least-element search on root element */
		/* bt_next = bt; */
		bt_next = e->root;
	}
	else {
		bt_next = NULL;

		/* dequeueing loop
		 *****************
		 * dequeue elements on stack until one of the following occurs:
		 *   - element's value has not been returned yet
		 *     (element has been enqueued during least-element search)
		 *   - element has a n_gt subtree which has not been processed yet
		 *     (<=> subtree's address != priviously dequeued element's)
		 *   - stack gets empty
		 *     (enumeration is completed, then)
		 */
		do {
#ifdef DEBUG
			zwarn("zbtree_enext(): dequeueing loop");
#endif

			/* stack size shrinkable by more than ZBTREE_SINC elements? */
			if ((e->stack_size - e->stack_depth) > ZBTREE_SINC) {
#ifdef DEBUG
				zwarn("zbtree_enext(): dequeueing loop: realloc()ing stack (%p)", e->stack);
#endif
				e->stack_size -= ZBTREE_SINC;
				if ((e->stack = realloc(e->stack,
				                        e->stack_size
							  * sizeof(struct zbtree *)))
				    == NULL) {
					zperror("zbtree_enext(): malloc()");
					return NULL;
				}
			}

			/* remember last element dequeued */
			bt_cur = bt_next;
			
			/* examine last element on stack */
			bt_next = e->stack[e->stack_depth - 1];
#ifdef DEBUG
			zwarn("zbtree_enext(): dequeueing loop: stack indexing ok: bt_next == %p", bt_next);
#endif

			/* returning from branch to lesser element?
			 * => element itself has not been returned yet
			 */
			if (bt_next->n_le && (bt_next->n_le == bt_cur))
				return bt_next;
			
			/* continue loop (see above)
			 ***************************
			 * dequeue last element on stack
			 * if previous loop continuation conditions satisfied
			 */
		} while ((!bt_next->n_gt || (bt_next->n_gt == bt_cur)) &&
		         (--(e->stack_depth) > 0));
		
		/* last element reached? */
		if (e->stack_depth == 0) {
#ifdef DEBUG
			zwarn("zbtree_enext(): dequeueing loop: last element reached");
#endif
			/* do cleanup: deallocate stack */
			e->stack = realloc(e->stack, 0);	
			e->stack_size = 0;
			
			/* return NULL to indicate end of enumeration */
			return NULL;
		}

		/* branch to previously unprocessed greater-than subtree */
		bt_next = bt_next->n_gt;
	}

	/* enqueueing loop
	 *****************
	 * do { ... } until there's no n_le / lesser value
	 */
	do {
#ifdef DEBUG
		zwarn("zbtree_enext(): enqueueing loop");
#endif
		
		/* continue with next element (see above & below) */
		bt_cur = bt_next;
		
		/* need to malloc() bigger stack? */
		if (e->stack_depth == e->stack_size) {
#ifdef DEBUG
			zwarn("zbtree_enext(): enqueueing loop: realloc()ing stack (%p)", e->stack);
#endif
			e->stack_size += ZBTREE_SINC;
			if ((e->stack = (struct zbtree **)
			                  realloc(e->stack,
				                  e->stack_size
						    * sizeof(struct zbtree *)))
			    == NULL) {
				zperror("zbtree_enext(): malloc()");
				return NULL;
			}
		}
		
		/* put bt_cur on the stack
		 *************************
		 * Each element gets put onto the stack at least once:
		 * when it's value is returned. After that,
		 * there's still it's n_gt... (being processed by next call)
		 */
#ifdef DEBUG
		zwarn("zbtree_enext(): enqueueing loop: putting (bt_cur == %p) onto the stack at %i", bt_cur, e->stack_depth);
#endif
		e->stack[e->stack_depth++] = bt_cur;

		/* continue with next element lesser than this one */
		bt_next = bt_cur->n_le;
	} while (bt_next);  /* stop when subtree's least element found */

	/* bt_cur now is on the stack and has no lesser element */
	return bt_cur;
}

/* zbtree_ereset -- reset zbtree_enum */
int zbtree_ereset(struct zbtree_enum *e)
{
	if (!e) {
		zerror("zbtree_ereset(): called with NULL pointer");
		return -1;
	}
	if (e->stack) {
		e->stack = realloc(e->stack, 0);
		e->stack_size = 0;
	}
	e->stack_depth = 0;
	return 0;
}

/* zbtree_eend() -- called after an enumeration to free up zbtree_enum */
struct zbtree_enum *zbtree_eend(struct zbtree_enum *e)
{
	if (!e) {
#if defined(DEBUG)
		zwarn("zbtree_eend(): not freeing NULL pointer");
#endif
		return NULL;
	}
	free(e->stack);
	return realloc(e, 0);
}
