/*
 * ----------------------------------------------
 *
 * This software belongs to the zzzutils package.
 *
 * Copyright (c) 2003 Fabian "zzznowman" Pietsch.
 * All rights reserved.
 *
 * ----------------------------------------------
 */

/*
 * zzzupdate-af.c -- search filesystem for audio files & put them into a database
 *
 ******************************************************************************
 *
 * <s>brief</s> ChangeLog:
 *   2003-05-05 original implementation as shell script
 *   2003-08-06 major changes & enhancements
 *   2003-08-06 reimplementation in C
 *   2003-08-08 finished zbtree implementation - now in zbtree.{c,h}
 *   2003-08-14 change chdir()-ing to ".." to chdir()-ing back to old cwd
 *   2003-08-14 directory recursion; fork of development as zzzsearchdir
 *   2003-08-15 match files with xmms media file extensions
 *   2003-08-16 fork(), execvp() xmms -e xmms_media_file_name
 *   2003-08-16 added variable controlling verbosity
 *   2003-08-16 overal code cleanup
 *   2003-08-18 changes are kept in file ChangeLog from now on
 *
 ******************************************************************************
 *
 *
 */

#include "zzzutils.h"
#include "zbtree.h"
#include "zzzupdate-af.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>


/* basename for source-of-output indication
 * NOTE: currently only used for usage information in usage()
 */
char *basename;

/* initial directories (cf. argc, argv) */
int initdirc = 1;
static char *const initdirv[] = {
	ZUAF_INITDIR
};
char *const *initdirvp = initdirv;	/* points to first directory argument, if any */

/* initial extension list */
int initextc = 2;
static char *const initextv[] = {
	".mp3",
	".ogg"
};
char *const *initextvp = initextv;

/* file/path exclusion list (if set, its last element has to be a NULL pointer) */
#if 0
static char *const exclv[] = {
	"www",
	NULL
};
char *const *exclvp = exclv;
#else
char *const *exclvp = NULL;
#endif

/* verbosity level for status & errors */
int verbose = 0;


/* combine path -- combine two partial path names */
char *combinepath(const char *path1, const char *path2)
{
	char *path;				/* buffer for combined path	*/
	size_t path1_len = strlen(path1);	/* length of path1		*/
	
	/* allocate buffer for combined path */
	if ((path = calloc(path1_len + strlen(path2) + 2,
			   sizeof(char))) == NULL) {
		zperror("combinepath(): malloc()");
		return NULL;
	}
	
	/* put together path: "path1" PATH_SEPARATOR "path2" */
	strcpy(path, path1);
	path[path1_len] = PATH_SEPARATOR;
	return strcat(path, path2);
}

/* is excluded -- does name match against file/path exclusion list? */
int isexcl(const char *name)
{
	char *const *excl = exclvp;	/* current element in exclusion list */
	
	/* exclusion list set? are there any elements in it? */
	if ((excl) && (*excl)) {
		/* enumerate through exclusion list
		 * while final NULL pointer isn't reached yet
		 */
		do {
			/* does name match exclusion list element? */
			if (strcmp(name, *excl) == 0)
				return 1;	/* is excluded */
		} while (*(++excl));
	}

	/* no match against exclusion list */
	return 0;	/* is not excluded */
}

/* is media file -- compare name's extension to those in the extension list */
int ismediafile(const char *name)
{
	int i, name_len, ext_len;  /* index in initextvp; strlen() of name, ext */

	/* get name's length */
	name_len = strlen(name);

	/* enumerate through list of valid extensions */
	for (i = 0; i < initextc; i++) {
		/* get currently compared extension's length */
		ext_len = strlen(initextvp[i]);

		/* name long enough to contain current extension?
		 * compare last chars to current extension, then (case insensitive)
		 */
		if ((ext_len < name_len) &&
		    (strcasecmp(&name[name_len - ext_len], initextvp[i]) == 0))
			return 1;	/* is media file */
	}

	/* no match against extension list */
	return 0;	/* is no media file */
}


/*
 * strlist_* functions -- manipulate NULL-terminated array of pointers to char
 *****************************************************************************
 * realloc() is used for resizing the array,
 * so it should have been allocated with malloc()
 */

/* FIXME: move this to a header file */
size_t strlist_len(const char **list);

/* strlist_add -- add string to string list */
int strlist_add(const char ***list, const char *str)
{
	const char **list_new;			/* list after realloc()	*/
	size_t len = strlist_len(*list);	/* list's length	*/
	
	/* reallocate list: old length + new element + NULL */
	if ((list_new = realloc(*list, (++len + 1) * sizeof(const char *)))
	    == NULL) {
		zperror("strlist_add(): malloc()");
		return -2;
	}

	/* append str to end of list */
	list_new[len - 1] = str;

	/* finish list with NULL pointer */
	list_new[len] = NULL;

	/* give pointer to new list back to the caller */
	*list = list_new;

	return 0;
}

/* strlist_len -- calculate list's length, not including the terminating NULL pointer */
size_t strlist_len(const char **list)
{
	size_t len = 0;		/* store length information */

	/* is list initialized? */
	if (list)
		/* loop through list & count elements */
		while (*(list++))
			len++;

	/* return list's length */
	return len;
}


/* search directory -- recursively search name and its subdirectories */
static int searchdir(const char *name)
{
	struct stat dir_info,		/* stat() for name,		*/
	            ent_info;		/*            dir entry		*/
	DIR *dir;			/* pointer to directory stream	*/
	struct dirent *dirent;		/* directory entry		*/
	struct zbtree *dir_dirs=NULL,	/* directory's subdirs,		*/
	              *dir_files=NULL,	/*             files		*/
		      *bt_cur;		/* zbtree enumeration		*/
	struct zbtree_enum *zbte;	/* enumerator for zbtree_e*	*/
	char *cwd_old;			/* absolute path to current wd	*/
	char *name_new;			/* subdir (for recursion)	*/
	int sd_ret,			/* used for searchdir recursion	*/
	    mfiles = 0;			/* count of media files found	*/

	
	/*
	 * check validity of the call
	 */
	
	/* stat name (for presence & directory check) */
	if (stat(name, &dir_info) != 0) {
		zperror("searchdir(): can't stat %s", name);
		return -1;
	}
	
	/* is name a directory? */
	if (!S_ISDIR(dir_info.st_mode)) {
		zerror("searchdir(): %s is not a directory", name);
		return -1;
	}


	/*
	 * do the real scan
	 */

	if (verbose >= 2)
		printf("%s: scanning %s...\n", basename, name);
	
	/* open directory stream */
	if ((dir = opendir(name)) == NULL) {
		zperror("searchdir(): opendir(\"%s\")", name);
		return -1;
	}

	/* remember current directory to return there later */
	if ((cwd_old = getcwd(NULL, 0)) == NULL) {
		zperror("searchdir(): getcwd()");
		return -2;
	}
	
	/* enter directory */
	if (chdir(name) != 0) {
		zperror("searchdir(): couldn't chdir() to %s", name);
		free(cwd_old);
		closedir(dir);
		return -1;
	}
	
	/* loop through directory entries */
	while ((dirent = readdir(dir))) {
		/* current (.) or parent (..) directory? */
		if (dirent->d_name[0] == '.') {
			switch (dirent->d_name[1]) {
			case 0:
			case '.':
				continue;
			}
		}
		
		/* stat directory entry */
		if (stat(dirent->d_name, &ent_info) != 0) {
			zperror("searchdir(): can't stat %s", dirent->d_name);
			continue;
		}
		
		/* is d_name a directory? (put into zbtree dir_dirs, then) */
		if (S_ISDIR(ent_info.st_mode))
			zbtree_add(&dir_dirs, strdup(dirent->d_name));
		/* is d_name a regular file? (put into zbtree dir_files, then) */
		else if (S_ISREG(ent_info.st_mode))
			zbtree_add(&dir_files, strdup(dirent->d_name));
	}
	dirent = NULL;

	/* leave directory */
	if (chdir(cwd_old) != 0) {
		zperror("searchdir(): couldn't chdir() back to %s", cwd_old);
		free(cwd_old);
		closedir(dir);
		zbtree_free(&dir_dirs, 1);
		zbtree_free(&dir_files, 1);
		return -2;
	}
	
	/* free absolute path to previous working directory */
	cwd_old = realloc(cwd_old, 0);
	
	/* close directory stream */
	if (closedir(dir) != 0)
		zperror("searchdir(): closedir()");


	/*
	 * process gathered information
	 */

	/* any files found? */
	if (dir_files) {
		/* start zbtree enumeration */
		if ((zbte = zbtree_estart(dir_files)) == NULL) {
			zerror("searchdir(): can't enumerate through zbtree dir_files");
			zbtree_free(&dir_files, 1);
			zbtree_free(&dir_dirs, 1);
			return -2;
		}

		/* walk through dir_files, automatically sorted by b-tree */
		while ((bt_cur = zbtree_enext(zbte))) {
			/* is this a media file? */
			if (ismediafile(bt_cur->v)) {
				/* put together path: "thisdir/mediafile" */
				if ((name_new = combinepath(name, bt_cur->v)) == NULL) {
					if (verbose >= 1)
						zerror("searchdir(): combinepath() failed");
					zbtree_free(&dir_dirs, 1);
					zbtree_free(&dir_files, 1);
					return -2;
				}

				/* check against exclusion list */
				if (isexcl(name_new) || isexcl(bt_cur->v)) {
					if (verbose >= 2)
						printf("%s: skipping file %s\n", basename, name_new);
					name_new = realloc(name_new, 0);
					continue;	/* skip excluded file */
				}

				/* process the media file */
				if (verbose >= 1)
					printf("%s: processing %s\n", basename, name_new);
				if (saveaf(name_new) < -1) {
					if (verbose >= 1)
						zerror("searchdir(): saveaf() failed in a non-recoverable way");
					return -2;
				}
				mfiles++;

				/* free buffer for media file's path */
				name_new = realloc(name_new, 0);
			}
		}

		/* end zbtree enumeration */
		zbte = zbtree_eend(zbte);
	}

	/* output progress information */
	if (mfiles && (verbose >= 0))
		printf("%s: processed %i files\n", name, mfiles);


	/*
	 * recurse into subdirectories
	 */

	/* any directories found? */
	if (dir_dirs) {
		/* start zbtree enumeration */
		if ((zbte = zbtree_estart(dir_dirs)) == NULL) {
			zerror("searchdir(): can't enumerate through zbtree dir_dirs");
			zbtree_free(&dir_files, 1);
			zbtree_free(&dir_dirs, 1);
			return -2;
		}

		/* walk through dir_dirs, automatically sorted by b-tree */
		while ((bt_cur = zbtree_enext(zbte))) {
			/* put together path: "thisdir/subdir" */
			if ((name_new = combinepath(name, bt_cur->v)) == NULL) {
				if (verbose >= 1)
					zerror("searchdir(): combinepath() failed");
				zbtree_free(&dir_dirs, 1);
				zbtree_free(&dir_files, 1);
				return -2;
			}

			/* check against exclusion list */
			if (isexcl(name_new) || isexcl(bt_cur->v)) {
				if (verbose >= 2)
					printf("%s: skipping directory %s\n", basename, name_new);
				name_new = realloc(name_new, 0);
				continue;	/* skip excluded directory */
			}

			/* recurse into subdirectory
			 * return with recursed searchdir()'s exit status
			 * in case of non-recoverable error
			 */
			if ((sd_ret = searchdir(name_new)) < -1)
				return sd_ret;

			/* free buffer for subdirectory's path */
			name_new = realloc(name_new, 0);
		}

		/* end zbtree enumeration */
		zbte = zbtree_eend(zbte);
	}


	/*
	 * do cleanup
	 */
	
	/* free used zbtrees
	 * (including dirent.d_name string copies malloc()ed by strdup())
	 */
	zbtree_free(&dir_dirs, 1);
	zbtree_free(&dir_files, 1);
	
	return 0;
}

/* usage -- output usage information; output to stderr & exit if doexit != 0 */
static void usage(int doexit)
{
	/* prepare usage message */
	char *msg = "\
Usage: %s [OPTION]... [DIRECTORY]\n"
PRODUCT_VERSION ": zzzupdate-af -- search DIRECTORY (or " ZUAF_INITDIR ", if none given)\n\
for audio files and process them: keep \"Project Stream\"'s database up to date\n\
Options:\n\
  -h     help    -- output this usage information and exit\n\
  -q     quiet   -- decrease verbosity\n\
  -v     verbose -- increase verbosity\n\
  -V     version -- output version information and exit\n\
  -x     eXclude -- add to file/directory exclusion list\n";

	/* output to stderr & exit? */
        if (doexit) {
                fprintf(stderr, msg, basename);
                exit(EX_USAGE);
        }
	
	/* output to stdout */
	printf(msg, basename);
}

/* main -- process arguments, start searchdir() recursions */
int main(int argc, char *const argv[])
{
	int i;			/* returned by getopt() & loop counter	*/
	
	/* set stdout linebuffered manually to avoid stderr strangely mixing in
	 * at unexpected places when both streams are the same (by redirection)
	 */
	setvbuf(stdout, NULL, _IOLBF, 0);

	/* get basename by stripping off leading path components */
	if ((basename = strrchr(argv[0], PATH_SEPARATOR)) == NULL)
		basename = argv[0];
	else
		basename++;

	/* process command line options */
	while ((i = getopt(argc, argv, "hqvVx:")) != -1) {
		switch (i) {
		case '?':	/* unrecognized option */
			/* output usage information (to stderr) & fail */
			usage(1);
			break;
		case 'h':	/* usage information requested */
			/* output usage information (to stdout) & succeed */
			usage(0);
			exit(EX_OK);
			break;
		case 'q':	/* decrease verbosity */
			verbose--;
			break;
		case 'v':	/* increase verbosity */
			verbose++;
			break;
		case 'V':	/* version information requested */
			printf("%s: version " PRODUCT_VERSION "\n", basename);
			exit(EX_OK);
			break;
		case 'x':	/* add to exclusion list */
			/* add optarg to exclusion list */
			if (strlist_add((const char ***) &exclvp, optarg) != 0) {
				/* adding to exclusion list failed */
				if (verbose >= 1)
					zerror("couldn't add %s to exclusion list", optarg);
				exit(EX_OSERR);
			}

			if (verbose >= 2)
				printf("%s: added %s to exclusion list\n", basename, optarg);
			break;
		default:	/* this should never happen */
			zerror("getopt() returned unexpected value 0%o -- exiting", i);
			exit(EX_SOFTWARE);
			break;
		}
	}
	
	/* any directories given on command line? */
	if (optind < argc) {
		/* override default initial directories */
		initdirvp = &argv[optind];
		initdirc = argc - optind;
	}

	/* initialize saveaf environment */
	if (saveaf_init()) {
		if (verbose >= 1)
			zerror("saveaf_init() failed -- exiting");
		exit(EX_OSERR);
	}

	/* start searchdir() recursion for each initial directory */
	for (i = 0; i < initdirc; i++) {
		if (searchdir(initdirvp[i]) < -1) {
			/* something went wrong */
			if (verbose >= 1)
				zerror("searchdir() failed in a non-recoverable way at %s -- exiting",
				       initdirvp[i]);

			/* fail */
			exit(EX_OSERR);
		}
	}

	/* deinitialize saveaf environment */
	if (saveaf_deinit()) {
		if (verbose >= 1)
			zerror("saveaf_deinit() failed -- exiting");
		exit(EX_OSERR);
	}

	/* succeed */
	exit(EX_OK);
}
