/*
 * ----------------------------------------------------------------------------
 *
 * zzzupdate-af_saveaf.c
 * Copyright (C) 2003 Fabian "zzznowman" Pietsch
 *
 * ----------------------------------------------------------------------------
 *
 */

#include "zzzutils.h"
#include "zbtree.h"
#include "zzzupdate-af.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mysql.h>


#define ZUAF_HOST	NULL
#define ZUAF_USER	NULL
#define ZUAF_DB		"user_fabian_stream"

#define ZUAF_SQL_DEL	"DELETE FROM audiofiles WHERE f_path='%s'"
#define ZUAF_SQL_GET	"SELECT f_path FROM audiofiles"
#define ZUAF_SQL_PUT	"INSERT INTO audiofiles (f_path) VALUES ('%s')"


/* base directory to check against and strip from name */
static char *basedir;
static int basedir_len;	/* increased efficiency */

/* already processed files */
static struct zbtree *processed = NULL;

/* MySQL client library data structures */
static MYSQL mysql;
static MYSQL_RES *mysqlres;


/* function prototypes */
int saveaf_init(void);
int saveaf(const char *name);
int saveaf_deinit(void);
static int saveaf_checkaf(const char *name);
static int saveaf_checkdb(void);
static int saveaf_isprocessed(const char *name);
static int saveaf_mysql_init(void);
static int saveaf_mysql_del(const char *name);
static int saveaf_mysql_get_init(void);
static char *saveaf_mysql_get(void);
static int saveaf_mysql_get_deinit(void);
static int saveaf_mysql_put(const char *name);


/* initialize saveaf environment: connect to MySQL server */
int saveaf_init() {
	int initdir_len = strlen(initdirvp[0]);

	/* exactly one target directory? */
	if (initdirc != 1) {
		zerror("there has to be _exactly_ one target directory (because of basedir stripping)");
		return -2;
	}

	/* allocate buffer for basedir */
	if ((basedir = malloc(sizeof(char) * (initdir_len + 2))) == NULL) {
		zperror("saveaf_init(): malloc()");
		return -2;
	}

	/* initialize basedir (no str_n_cpy needed, size checked through malloc()) */
	strncpy(basedir, initdirvp[0], initdir_len + 2);

	/* check for trailing slash */
	if (basedir[initdir_len - 1] != '/') {
		/* add slash; strncpy (hopefully!) padded basedir with nulls */
		basedir[initdir_len] = '/';
	}

	/* increase efficiency by reducing strlen() usage */
	basedir_len = strlen(basedir);

	/* initialize MySQL environment */
	if (saveaf_mysql_init()) {
		if (verbose >= 1)
			zerror("saveaf_init(): saveaf_mysql_init() failed");
		return -2;
	}

	/* validate existing database */
	if (saveaf_checkdb() < 0) {
		if (verbose >= 1)
			zerror("saveaf_init(): saveaf_checkdb() failed");
		return -2;
	}

	/* success */
	return 0;
}

/* save af -- save audiofile: store information about audio file in MySQL db */
int saveaf(const char *name)
{
	int ret;

	/* does name begin with base directory? */
	if (strncmp(name, basedir, basedir_len)) {
		zwarn("saveaf(): rejecting %s: not below base dir", name);
		return -1;
	}

	/* strip base directory */
	name += basedir_len;

	/* audio file already processed? */
	if (saveaf_isprocessed(name))
		return 0;

	/* put entry into database */
	if ((ret = saveaf_mysql_put(name)) < -1) {
		if (verbose >= 1)
			zerror("saveaf(): saveaf_mysql_put(\"name\") failed in a non-recoverable way", name);
		return -2;
	}
	else if (ret)
		return -1;

	/* mark audio file as being already processed */
	if (zbtree_add(&processed, strdup(name)) == NULL)
		return -2;

	/* success */
	return 0;
}

/* saveaf environment deinitialization: close connection to MySQL server */
int saveaf_deinit() {
	/* close connection to MySQL server */
	mysql_close(&mysql);

	/* free zbtree holding already processed audio files */
	zbtree_free(&processed, 1);

	/* free base directory name */
	free(basedir);

	/* success */
	return 0;
}


/* check audio file -- check if data for this af-db entry is still correct */
static int saveaf_checkaf(const char *name) {
	char *name_full;
	int name_full_len = basedir_len + strlen(name) + 1;
	struct stat file;

	/* malloc() name_full */
	if ((name_full = malloc(sizeof(char) * name_full_len)) == NULL) {
		zperror("saveaf_checkaf(): malloc()");
		return -2;
	}

	/* prepend base directory to name */
	strcpy(name_full, basedir);
	strcat(name_full, name);

	/* stat file (to check for its presence) */
	if (stat(name_full, &file)) {
		if (verbose >= 2)
			zperror("saveaf_checkaf(\"%s\"): stat()", name);
		free(name_full);

		/* data has changed */
		return 1;
	}
	free(name_full);

	if (!S_ISREG(file.st_mode)) {
		/* data has changed */
		return 1;
	}

	/* success (data is still correct) */
	return 0;
}

/* validate existing audio files database */
static int saveaf_checkdb() {
	int rows, uptodate = 0, check;
	char *name;

	/* query for all audio files in database */
	if ((rows = saveaf_mysql_get_init()) < 0) {
		zerror("saveaf_checkdb(): saveaf_mysql_get_init() failed");
		return -2;
	}

	/* be verbose about what's going on */
	if (verbose >= 0)
		printf("checking %lld current entries in audio file database...\n", mysql_num_rows(mysqlres));

	/* retrieve audio file database file by file */
	while ((name = saveaf_mysql_get())) {
		/* check stored data about audio file */
		if ((check = saveaf_checkaf(name)) < 0) {
			zerror("saveaf_checkdb(): saveaf_checkaf() failed");
			return -2;
		}
		else if (check == 0) {
			if (verbose >= 1)
				printf("%s: up to date\n", name);
			uptodate++;

			/* mark as being already processed */
			zbtree_add(&processed, strdup(name));
		}
		else {
			if (verbose >= 1)
				printf("%s: outdated, removing entry\n", name);

			/* delete row from database */
			if (saveaf_mysql_del(name)) {
				zerror("saveaf_checkdb(): saveaf_mysql_del(\"%s\") failed", name);
				return -2;
			}
		}

		/* free memory allocated by strdup() in saveaf_mysql_get() */
		free(name);
	}

	/* be verbose about what's going on */
	if (verbose >= 0)
		printf("check complete: %d of %d values up to date\n", uptodate, rows);

	/* deinitialize audio file query */
	if (saveaf_mysql_get_deinit()) {
		zerror("saveaf_checkdb(): saveaf_mysql_deinit() failed");
		return -2;
	}

	/* success */
	return 0;
}

/* has audio file already been processed? */
static int saveaf_isprocessed(const char *name) {
	struct zbtree *bt = processed;
	int cmp;

	/* loop while relevant subtree left */
	while (bt) {
		if ((cmp = strcoll(name, bt->v)) == 0)
			/* equal => audio file already processed */
			return 1;
		else if (cmp > 0)
			/* greater than => use next greater-than node */
			bt = bt->n_gt;
		else
			/* less than => use next less-or-equal node */
			bt = bt->n_le;
	}

	/* audio file not yet processed */
	return 0;
}


/*
 * MySQL related functions
 */

/* initialize MySQL environment */
static int saveaf_mysql_init() {
	int i;
	char *password;

	/* initialize MySQL object */
	if (mysql_init(&mysql) == NULL) {
		zerror("saveaf_mysql_init(): mysql_init(): %s", mysql_error(&mysql));
		return -2;
	}
	
	/* read in my.cnf */
	mysql_options(&mysql, MYSQL_READ_DEFAULT_GROUP, "");

	/* input MySQL password */
	password = get_tty_password(NULL);

	/* connect to MySQL server */
	if (mysql_real_connect(&mysql, ZUAF_HOST, ZUAF_USER, password,
	      ZUAF_DB, 0, NULL, 0) == NULL) {
		zerror("saveaf_mysql_init(): mysql_real_connect(): %s", mysql_error(&mysql));
		return -2;
	}
	
	/* clear password */
	for (i=0; i<strlen(password); i++)
		password[i] = '\0';

	/* success */
	return 0;
}

/* delete audio file from database */
static int saveaf_mysql_del(const char *name) {
	int name_len = strlen(name), query_len;
	char *name_esc, *query;
	my_ulonglong rows;

	/* allocate len*2 + 1 chars for mysql_real_escape_string */
	if ((name_esc = malloc(sizeof(char) * (name_len * 2 + 1))) == NULL) {
		zperror("saveaf_mysql_del(): malloc()");
		return -2;
	}

	/* escape name prior to using it in MySQL query
	 * according to MySQL documentation, there are no possible errors...
	 */
	mysql_real_escape_string(&mysql, name_esc, name, name_len);

	/* get size of buffer needed for query */
	if ((query_len = snprintf(NULL, 0, ZUAF_SQL_DEL, name_esc)) < 1) {
		/* this snprintf isn't friendly enough... => estimate roughly */
		query_len = 1000;
	}

	/* allocate buffer for query */
	if ((query = malloc(sizeof(char) * (query_len + 1))) == NULL) {
		zperror("saveaf_mysql_del(): malloc()");
		free(name_esc);
		return -2;
	}

	/* prepare query string */
	snprintf(query, query_len + 1, ZUAF_SQL_DEL, name_esc);
	free(name_esc);

	/* execute query */
	if (mysql_real_query(&mysql, query, strlen(query))) {
		zerror("saveaf_mysql_del(): mysql_real_query(): %s", mysql_error(&mysql));
		free(query);
		return -1;
	}
	free(query);

	/* check result */
	if ((rows = mysql_affected_rows(&mysql)) != 1) {
		zerror("saveaf_mysql_del(): mysql_affected_rows() didn't return 1 but %d", rows);
		return -1;
	}

	/* success */
	return 0;
}

/* query all audio files in database */
static int saveaf_mysql_get_init() {
	/* query MySQL */
	if (mysql_real_query(&mysql, ZUAF_SQL_GET, strlen(ZUAF_SQL_GET))) {
		zerror("saveaf_mysql_get_init(): mysql_real_query(): %s", mysql_error(&mysql));
		return -2;
	}

	/* prepare result retrieval */
	if ((mysqlres = mysql_store_result(&mysql)) == NULL) {
		zerror("saveaf_mysql_get_init(): mysql_store_result(): %s", mysql_error(&mysql));
		return -2;
	}

	/* success */
	return mysql_num_rows(mysqlres);
}

/* retrieve next row from result recordset for query for all audio files in database */
static char *saveaf_mysql_get() {
	MYSQL_ROW row;

	/* fetch next row from result set */
	if ((row = mysql_fetch_row(mysqlres)) == NULL) {
		/* no rows left */
		return NULL;
	}

	/* return strdup()ed version of first (and only) field in row */
	return strdup(row[0]);
}

/* deinitialize query for all audio files currently in database */
static int saveaf_mysql_get_deinit() {
	/* free the query's result set */
	mysql_free_result(mysqlres);

	/* success */
	return 0;
}

static int saveaf_mysql_put(const char *name) {
	int name_len = strlen(name), query_len;
	char *name_esc, *query;
	my_ulonglong rows;

	/* allocate len*2 + 1 chars for mysql_real_escape_string */
	if ((name_esc = malloc(sizeof(char) * (name_len * 2 + 1))) == NULL) {
		zperror("saveaf_mysql_put(): malloc()");
		return -2;
	}

	/* escape name prior to using it in MySQL query
	 * according to MySQL documentation, there are no possible errors...
	 */
	mysql_real_escape_string(&mysql, name_esc, name, name_len);

	/* get size of buffer needed for query */
	if ((query_len = snprintf(NULL, 0, ZUAF_SQL_PUT, name_esc)) < 1) {
		/* this snprintf isn't friendly enough... => estimate roughly */
		query_len = 1000;
	}

	/* allocate buffer for query */
	if ((query = malloc(sizeof(char) * (query_len + 1))) == NULL) {
		zperror("saveaf_mysql_put(): malloc()");
		free(name_esc);
		return -2;
	}

	/* prepare query string */
	snprintf(query, query_len + 1, ZUAF_SQL_PUT, name_esc);
	free(name_esc);

	/* execute query */
	if (mysql_real_query(&mysql, query, strlen(query))) {
		zerror("saveaf_mysql_put(): mysql_real_query(): %s", mysql_error(&mysql));
		free(query);
		return -1;
	}
	free(query);

	/* check result */
	if ((rows = mysql_affected_rows(&mysql)) != 1) {
		zerror("saveaf_mysql_put(): mysql_affected_rows() didn't return 1 but %d", rows);
		return -1;
	}

	/* success */
	return 0;
}
